"""Entry point for client creation.

``build_client_manager`` in particular is the abstraction that should be used
to create a ``ClientManager``, that in return can create Pulsar clients for
specific actions.
"""

import functools
import threading
from logging import getLogger
from os import getenv
from queue import Queue
from typing import (
    Any,
    Dict,
    Optional,
    Type,
    TYPE_CHECKING,
)

from typing_extensions import Protocol

from .amqp_exchange_factory import get_exchange
from .client import (
    BaseJobClient,
    InputCachingJobClient,
    JobClient,
    GcpMessageCoexecutionJobClient,
    GcpPollingCoexecutionJobClient,
    K8sMessageCoexecutionJobClient,
    K8sPollingCoexecutionJobClient,
    MessageCLIJobClient,
    MessageJobClient,
    RelayJobClient,
    TesMessageCoexecutionJobClient,
    TesPollingCoexecutionJobClient,
)
from .destination import url_to_destination_params
from .object_client import ObjectStoreClient
from .server_interface import (
    HttpPulsarInterface,
    LocalPulsarInterface,
    PulsarInterface,
)
from .transport import get_transport
from .transport.relay import RelayTransport
from .util import TransferEventManager

if TYPE_CHECKING:
    from pulsar.managers import ManagerInterface

log = getLogger(__name__)

DEFAULT_TRANSFER_THREADS = 2


class ClientManagerInterface(Protocol):

    def get_client(self, destination_params: Dict[str, Any], job_id: str, **kwargs: Dict[str, Any]) -> BaseJobClient:
        """Get client instance for specified job description."""

    def shutdown(self, ensure_cleanup=False) -> None:
        """Mark client manager's work as complete and clean up resources it managed."""
        return


class ClientManager(ClientManagerInterface):
    """Factory class to create Pulsar clients.

    This class was introduced for classes of clients that need to potential
    share state between multiple client connections.
    """
    job_manager_interface_class: Type[PulsarInterface]
    client_class: Type[BaseJobClient]

    def __init__(self, job_manager: Optional["ManagerInterface"] = None, **kwds: Dict[str, Any]):
        """Build a HTTP client or a local client that talks directly to a job manger."""
        if 'pulsar_app' in kwds or job_manager:
            self.job_manager_interface_class = LocalPulsarInterface
            pulsar_app = kwds.get('pulsar_app', None)
            file_cache = kwds.get('file_cache', None)
            self.job_manager_interface_args = dict(
                job_manager=job_manager,
                pulsar_app=pulsar_app,
                file_cache=file_cache,
            )
        else:
            print(kwds)
            self.job_manager_interface_class = HttpPulsarInterface
            transport_type = kwds.get('transport', None)
            transport_params = {p.replace('transport_', '', 1): v for p, v in kwds.items() if p.startswith('transport_')}
            transport = get_transport(transport_type, transport_params=transport_params)
            self.job_manager_interface_args = dict(transport=transport)
        cache = kwds.get('cache', None)
        if cache is None:
            cache = _environ_default_int('PULSAR_CACHE_TRANSFERS')
        if cache:
            log.info("Setting Pulsar client class to caching variant.")
            self.client_cacher = ClientCacher(**kwds)
            self.client_class = InputCachingJobClient
            self.extra_client_kwds = {"client_cacher": self.client_cacher}
        else:
            log.info("Setting Pulsar client class to standard, non-caching variant.")
            self.client_class = JobClient
            self.extra_client_kwds = {}

    def get_client(self, destination_params, job_id, **kwargs):
        """Build a client given specific destination parameters and job_id."""
        destination_params = _parse_destination_params(destination_params)
        destination_params.update(**kwargs)
        job_manager_interface_class = self.job_manager_interface_class
        job_manager_interface_args = dict(destination_params=destination_params, **self.job_manager_interface_args)
        job_manager_interface = job_manager_interface_class(**job_manager_interface_args)
        return self.client_class(destination_params, job_id, job_manager_interface, **self.extra_client_kwds)


try:
    from galaxy.jobs.runners.util.cli import factory as cli_factory
except ImportError:
    from pulsar.managers.util.cli import factory as cli_factory


class BaseRemoteConfiguredJobClientManager(ClientManagerInterface):

    def __init__(self, **kwds: Dict[str, Any]):
        self.manager_name = kwds.get("manager", None) or "_default_"


class MessageQueueClientManager(BaseRemoteConfiguredJobClientManager):
    status_cache: Dict[str, Any]
    ack_consumer_threads: Dict[str, threading.Thread]

    def __init__(self, amqp_url: str, **kwds: Dict[str, Any]):
        super().__init__(**kwds)
        self.url = amqp_url
        self.amqp_key_prefix = kwds.get("amqp_key_prefix", None)
        self.exchange = get_exchange(self.url, self.manager_name, kwds)
        self.status_cache = {}
        self.callback_lock = threading.Lock()
        self.callback_thread = None
        self.ack_consumer_threads = {}
        self.active = True

    def callback_wrapper(self, callback, body, message):
        if message.acknowledged:
            log.info("Message is already acknowledged (by an upstream "
                     "callback?), Pulsar client will not handle this message")
            return
        if not self.active:
            log.debug("Obtained update message for inactive client manager, attempting requeue.")
            try:
                message.requeue()
                log.debug("Requeue succeeded, will likely be handled next time consumer is enabled.")
            except Exception:
                log.debug("Requeue failed, message may be lost?")
            return

        try:
            if "job_id" in body:
                job_id = body["job_id"]
                self.status_cache[job_id] = body
            log.debug("Handling asynchronous status update from remote Pulsar.")
            callback(body)
        except Exception:
            log.exception("Failure processing job status update message.")
        except BaseException as e:
            log.exception("Failure processing job status update message - BaseException type %s" % type(e))
        finally:
            message.ack()

    def callback_consumer(self, callback_wrapper):
        try:
            self.exchange.consume("status_update", callback_wrapper, check=self)
        except Exception:
            log.exception("Exception while handling status update messages, "
                          "this shouldn't really happen. Handler should be "
                          "restarted.")
        finally:
            log.debug("Leaving Pulsar client status update thread, no "
                      "additional Pulsar updates will be processed.")

    def ensure_has_status_update_callback(self, callback):
        with self.callback_lock:
            if self.callback_thread is not None:
                return

            callback_wrapper = functools.partial(self.callback_wrapper, callback)
            run = functools.partial(self.callback_consumer, callback_wrapper)
            thread = threading.Thread(
                name="pulsar_client_%s_status_update_callback" % self.manager_name,
                target=run
            )
            thread.daemon = False  # Lets not interrupt processing of this.
            thread.start()
            self.callback_thread = thread

    def ack_consumer(self, queue_name: str):
        try:
            self.exchange.consume(queue_name + '_ack', None, check=self)
        except Exception:
            log.exception("Exception while handling %s acknowledgement "
                          "messages, this shouldn't really happen. Handler "
                          "should be restarted.", queue_name)
        finally:
            log.debug("Leaving Pulsar client %s acknowledgement thread, no "
                      "additional acknowledgements will be processed.",
                      queue_name)

    def ensure_has_ack_consumers(self):
        with self.callback_lock:
            for name in ('setup', 'kill'):
                if name in self.ack_consumer_threads:
                    return

                run = functools.partial(self.ack_consumer, name)
                thread = threading.Thread(
                    name="pulsar_client_{}_{}_ack".format(self.manager_name, name),
                    target=run
                )
                thread.daemon = False  # Lets not interrupt processing of this.
                thread.start()
                self.ack_consumer_threads[name] = thread

    def shutdown(self, ensure_cleanup: bool = False):
        self.active = False
        if ensure_cleanup:
            if self.callback_thread is not None:
                self.callback_thread.join()
            for v in self.ack_consumer_threads.values():
                v.join()

    def __nonzero__(self):
        return self.active

    __bool__ = __nonzero__  # Both needed Py2 v 3

    def get_client(self, destination_params, job_id, **kwargs):
        if job_id is None:
            raise Exception("Cannot generate Pulsar client for empty job_id.")
        destination_params = _parse_destination_params(destination_params)
        destination_params.update(**kwargs)
        if self.amqp_key_prefix:
            destination_params["amqp_key_prefix"] = self.amqp_key_prefix
        if 'shell_plugin' in destination_params:
            shell = cli_factory.get_shell(destination_params)
            return MessageCLIJobClient(destination_params, job_id, self, shell)
        elif destination_params.get('k8s_enabled', False):
            return K8sMessageCoexecutionJobClient(destination_params, job_id, self)
        elif destination_params.get("tes_url", False):
            return TesMessageCoexecutionJobClient(destination_params, job_id, self)
        elif destination_params.get("project_id", False):
            return GcpMessageCoexecutionJobClient(destination_params, job_id, self)
        else:
            return MessageJobClient(destination_params, job_id, self)


class RelayClientManager(BaseRemoteConfiguredJobClientManager):
    """Client manager that communicates with Pulsar via pulsar-relay.

    This manager uses HTTP-based long-polling to receive status updates from
    Pulsar through the relay, while posting control messages (setup, status
    requests, kill) to the relay for Pulsar to consume.
    """
    status_cache: Dict[str, Any]

    def __init__(self, relay_url: str, relay_username: str, relay_password: str, relay_topic_prefix: str = '', **kwds: Dict[str, Any]):
        super().__init__(**kwds)

        if not relay_url:
            raise Exception("relay_url is required for RelayClientManager")

        # Initialize relay transport
        self.relay_transport = RelayTransport(relay_url, relay_username, relay_password)
        self.relay_topic_prefix = relay_topic_prefix
        self.status_cache = {}
        self.callback_lock = threading.Lock()
        self.callback_thread = None
        self.active = True
        self.shutdown_event = threading.Event()

    def callback_wrapper(self, callback, message_data):
        """Process status update messages from the relay."""
        if not self.active:
            log.debug("Obtained update message for inactive client manager, ignoring.")
            return

        try:
            payload = message_data.get('payload', {})
            if "job_id" in payload:
                job_id = payload["job_id"]
                self.status_cache[job_id] = payload
            log.debug("Handling asynchronous status update from Pulsar via relay.")
            callback(payload)
        except Exception:
            log.exception("Failure processing job status update message.")
        except BaseException as e:
            log.exception("Failure processing job status update message - BaseException type %s" % type(e))

    def status_consumer(self, callback_wrapper):
        """Long-poll the relay for status update messages."""
        manager_name = self.manager_name
        topic = self._make_topic_name("job_status_update", manager_name)

        log.info("Starting relay status consumer for topic '%s'", topic)

        while self.active:
            try:
                # Long poll for status updates (30 second timeout)
                messages = self.relay_transport.long_poll([topic], timeout=30)

                for message in messages:
                    callback_wrapper(message)

            except Exception:
                if self.active:
                    log.exception("Exception while polling for status updates from relay, will retry.")
                    # Brief sleep before retrying to avoid tight loop on persistent errors
                    # Use wait() instead of sleep() to allow immediate interruption on shutdown
                    if self.shutdown_event.wait(timeout=5):
                        break
                else:
                    log.debug("Exception during shutdown, ignoring.")
                    break

        log.info("Done consuming relay status updates for topic %s", topic)

    def ensure_has_status_update_callback(self, callback):
        """Start a thread to poll for status updates if not already running."""
        with self.callback_lock:
            if self.callback_thread is not None:
                return

            callback_wrapper = functools.partial(self.callback_wrapper, callback)
            run = functools.partial(self.status_consumer, callback_wrapper)
            thread = threading.Thread(
                name="pulsar_client_%s_relay_status_consumer" % self.manager_name,
                target=run
            )
            # Make daemon so Python can exit even if thread is blocked in HTTP request.
            # Unlike MessageQueueClientManager which uses AMQP connections that can be
            # interrupted cleanly, HTTP long-poll requests block until timeout.
            thread.daemon = True
            thread.start()
            self.callback_thread = thread

    def ensure_has_ack_consumers(self):
        """No-op for relay client manager, as acknowledgements are handled via HTTP."""
        pass

    def _make_topic_name(self, base_topic: str, manager_name: str) -> str:
        """Create a topic name with optional prefix and manager suffix.

        Args:
            base_topic: Base topic name (e.g., 'job_setup', 'job_status_update')
            manager_name: Manager name (e.g., '_default_', 'cluster_a')

        Returns:
            Fully qualified topic name
        """
        parts = []

        # Add prefix if provided
        if self.relay_topic_prefix:
            parts.append(self.relay_topic_prefix)

        # Add base topic
        parts.append(base_topic)

        # Add manager name if not default
        if manager_name != "_default_":
            parts.append(manager_name)

        return "_".join(parts)

    def shutdown(self, ensure_cleanup: bool = False):
        """Shutdown the client manager and cleanup resources."""
        self.active = False
        # Signal the shutdown event to interrupt any waiting threads
        self.shutdown_event.set()
        if ensure_cleanup:
            if self.callback_thread is not None:
                self.callback_thread.join()
        # Close relay transport
        if hasattr(self, 'relay_transport'):
            self.relay_transport.close()

    def __nonzero__(self):
        return self.active

    __bool__ = __nonzero__  # Both needed Py2 v 3

    def get_client(self, destination_params, job_id, **kwargs):
        """Create a RelayJobClient for the given job."""
        if job_id is None:
            raise Exception("Cannot generate Pulsar client for empty job_id.")
        destination_params = _parse_destination_params(destination_params)
        destination_params.update(**kwargs)
        return RelayJobClient(destination_params, job_id, self)


class PollingJobClientManager(BaseRemoteConfiguredJobClientManager):

    def get_client(self, destination_params, job_id, **kwargs):
        if job_id is None:
            raise Exception("Cannot generate Pulsar client for empty job_id.")
        destination_params = _parse_destination_params(destination_params)
        destination_params.update(**kwargs)
        # TODO: cli version of this...
        if destination_params.get('k8s_enabled', False):
            return K8sPollingCoexecutionJobClient(destination_params, job_id, self)
        elif destination_params.get("tes_url", False):
            return TesPollingCoexecutionJobClient(destination_params, job_id, self)
        elif destination_params.get("project_id", False):
            return GcpPollingCoexecutionJobClient(destination_params, job_id, self)
        else:
            raise Exception("Unknown client configuration")

    def shutdown(self, ensure_cleanup=False):
        pass


def build_client_manager(
    job_manager: Optional["ManagerInterface"] = None,
    relay_url: Optional[str] = None,
    relay_username: Optional[str] = None,
    relay_password: Optional[str] = None,
    relay_topic_prefix: Optional[str] = None,
    amqp_url: Optional[str] = None,
    k8s_enabled: Optional[bool] = None,
    tes_enabled: Optional[bool] = None,
    gcp_batch_enabled: Optional[bool] = None,
    **kwargs
) -> ClientManagerInterface:
    if job_manager:
        return ClientManager(job_manager=job_manager, **kwargs)  # TODO: Consider more separation here.
    elif relay_url:
        assert relay_password and relay_username, "relay_url set, but relay_username and relay_password must also be set"
        return RelayClientManager(
            relay_url=relay_url,
            relay_username=relay_username,
            relay_password=relay_password,
            relay_topic_prefix=relay_topic_prefix or '',
            **kwargs
        )
    elif amqp_url:
        return MessageQueueClientManager(amqp_url=amqp_url, **kwargs)
    elif k8s_enabled or tes_enabled or gcp_batch_enabled:
        return PollingJobClientManager(**kwargs)
    else:
        return ClientManager(**kwargs)


class ObjectStoreClientManager:

    def __init__(self, **kwds):
        if 'object_store' in kwds:
            self.interface_class = LocalPulsarInterface
            self.interface_args = dict(object_store=kwds['object_store'])
        else:
            self.interface_class = HttpPulsarInterface
            transport_type = kwds.get('transport', None)
            transport = get_transport(transport_type)
            self.interface_args = dict(transport=transport)
        self.extra_client_kwds = {}

    def get_client(self, client_params):
        interface_class = self.interface_class
        interface_args = dict(destination_params=client_params, **self.interface_args)
        interface = interface_class(**interface_args)
        return ObjectStoreClient(interface)


class ClientCacher:

    def __init__(self, **kwds):
        self.event_manager = TransferEventManager()
        default_transfer_threads = _environ_default_int('PULSAR_CACHE_THREADS', DEFAULT_TRANSFER_THREADS)
        num_transfer_threads = int(kwds.get('transfer_threads', default_transfer_threads))
        self.__init_transfer_threads(num_transfer_threads)

    def queue_transfer(self, client, path):
        self.transfer_queue.put((client, path))

    def acquire_event(self, input_path):
        return self.event_manager.acquire_event(input_path)

    def _transfer_worker(self):
        while True:
            transfer_info = self.transfer_queue.get()
            try:
                self.__perform_transfer(transfer_info)
            except BaseException as e:
                log.warn("Transfer failed.")
                log.exception(e)
                pass
            self.transfer_queue.task_done()

    def __perform_transfer(self, transfer_info):
        (client, path) = transfer_info
        event_holder = self.event_manager.acquire_event(path, force_clear=True)
        failed = True
        try:
            client.cache_insert(path)
            failed = False
        finally:
            event_holder.failed = failed
            event_holder.release()

    def __init_transfer_threads(self, num_transfer_threads):
        self.num_transfer_threads = num_transfer_threads
        self.transfer_queue = Queue()
        for _ in range(num_transfer_threads):
            t = threading.Thread(target=self._transfer_worker)
            t.daemon = True
            t.start()


def _parse_destination_params(destination_params):
    try:
        unicode_type = unicode
    except NameError:
        unicode_type = str
    if isinstance(destination_params, str) or isinstance(destination_params, unicode_type):
        destination_params = url_to_destination_params(destination_params)
    return destination_params


def _environ_default_int(variable, default="0"):
    val = getenv(variable, default)
    int_val = int(default)
    if str(val).isdigit():
        int_val = int(val)
    return int_val


__all__ = (
    'ClientManager',
    'ObjectStoreClientManager',
    'HttpPulsarInterface',
)



import os
import re
from typing import Dict
from urllib.parse import urlparse

import requests

try:
    from tusclient import client
    tus_client_available = True
except ImportError:
    tus_client_available = False

TUS_CLIENT_UNAVAILABLE_MESSAGE = \
    "You are attempting to use the Tus transport with the Pulsar client but tuspy is unavailable."
DEFAULT_PULSAR_TUS_CHUNK_SIZE = 10**7
PULSAR_TUS_CHUNK_SIZE = int(os.getenv('PULSAR_TUS_CHUNK_SIZE', DEFAULT_PULSAR_TUS_CHUNK_SIZE))


def tus_upload_file(url: str, path: str) -> None:
    if not tus_client_available:
        raise Exception(TUS_CLIENT_UNAVAILABLE_MESSAGE)

    storage = None
    metadata: Dict[str, str] = {}

    headers: Dict[str, str] = {}
    tus_url = find_tus_endpoint(url)
    my_client = client.TusClient(tus_url, headers=headers)
    uploader = my_client.uploader(path, metadata=metadata, url_storage=storage)
    uploader.chunk_size = PULSAR_TUS_CHUNK_SIZE
    uploader.upload()
    upload_session_url = uploader.url
    assert upload_session_url
    tus_session_id = upload_session_url.rsplit("/", 1)[1]
    # job_key and such are encoded in the URL but this route expects a POST body
    # and if it has one Galaxy sticks the URL parameters into the "payload" object
    # for the controller. So encoded session_id in the POST body - it probably
    # all belongs there anyway.
    post_response = requests.post(url, data={"session_id": tus_session_id})
    post_response.raise_for_status()


def find_tus_endpoint(job_files_endpoint: str) -> str:
    parsed = urlparse(job_files_endpoint)
    job_files_url_path = parsed.path
    tus_endpoint = re.sub(r"jobs/[^/]*/files", "job_files/resumable_upload", job_files_url_path, 1)

    new_url = parsed._replace(path=tus_endpoint)
    return new_url.geturl()



""" CLI related utilities for submitting Pulsar jobs.
"""
import fnmatch
import sys
import uuid

from pulsar.client import (
    ClientJobDescription,
    ClientOutputs,
    finish_job,
    PulsarOutputs,
    submit_job,
)
from pulsar.client.test.check import (
    client_info,
    extract_client_options,
    HELP_DISABLE_CLEANUP,
    HELP_JOB_ID,
    HELP_PRIVATE_TOKEN,
    HELP_SUPPRESS_OUTPUT,
    HELP_TRANSPORT,
    HELP_URL,
    Waiter,
)
from pulsar.client.util import json_dumps
from pulsar.main import ArgumentParser
from pulsar.scripts.submit_util import (
    add_common_submit_args,
    run_server_for_job,
)

HELP_AMQP_URL = "Communicate with Pulsar listining on a message queue at this URL."
HELP_SERVER = "Run a Pulsar server locally instead of contacting a remote one."
HELP_COMMAND = "Shell command to execute on Pulsar server."
HELP_WORKING_DIRECTORY = "Local working directory (will be translated to a new directory)."
HELP_OUTPUT = "Output glob to collect from job (relative to remote working directory)."
HELP_OUTPUT_PATTERN = "Output pattern to collect from job (relative to remote working directory)."

DEFAULT_CLIENT_URL = 'http://localhost:8913/'


def main(argv=None):
    mod_docstring = sys.modules[__name__].__doc__
    arg_parser = ArgumentParser(description=mod_docstring)
    add_common_submit_args(arg_parser)
    arg_parser.add_argument('--url', default=DEFAULT_CLIENT_URL, help=HELP_URL)
    arg_parser.add_argument('--amqp_url', default=DEFAULT_CLIENT_URL, help=HELP_AMQP_URL)
    arg_parser.add_argument('--private_token', default=None, help=HELP_PRIVATE_TOKEN)
    # TODO: choices...
    arg_parser.add_argument('--default_file_action', default="none")
    arg_parser.add_argument('--file_action_config', default=None)
    arg_parser.add_argument('--transport', default=None, choices=["urllib", "curl"], help=HELP_TRANSPORT)  # set to curl to use pycurl
    arg_parser.add_argument('--suppress_output', default=False, action="store_true", help=HELP_SUPPRESS_OUTPUT)
    arg_parser.add_argument('--disable_cleanup', dest="cleanup", default=True, action="store_false", help=HELP_DISABLE_CLEANUP)
    arg_parser.add_argument('--server', default=False, action="store_true", help=HELP_SERVER)
    arg_parser.add_argument('--job_id', default=None, help=HELP_JOB_ID)
    arg_parser.add_argument('--command', help=HELP_COMMAND)
    arg_parser.add_argument('--working_directory', default=".", help=HELP_WORKING_DIRECTORY)
    arg_parser.add_argument('--result_json', default=None)
    arg_parser.add_argument('--output', default=[], action="append", help=HELP_OUTPUT)
    arg_parser.add_argument('--output_pattern', default=[], action="append", help=HELP_OUTPUT_PATTERN)

    args = arg_parser.parse_args(argv)
    if args.server:
        return run_server_for_job(args)
    else:
        failed = _run_client_for_job(args)
        if failed:
            return 1
        else:
            return 0


def _run_client_for_job(args):
    if args.job_id is None:
        args.job_id = str(uuid.uuid4())
    output_patterns = []
    output_patterns.extend(args.output_pattern)
    for output in args.output:
        output_patterns.append(fnmatch.translate(output))

    client_options = extract_client_options(args)
    client, client_manager = client_info(args, client_options)
    try:
        working_directory = args.working_directory
        client_outputs = ClientOutputs(
            working_directory=working_directory,
            dynamic_outputs=output_patterns,
        )
        job_description = ClientJobDescription(
            command_line=args.command,
            working_directory=working_directory,
            client_outputs=client_outputs,
        )
        submit_job(client, job_description)
        waiter = Waiter(client, client_manager)
        result_status = waiter.wait()
        pulsar_outputs = PulsarOutputs.from_status_response(result_status)
        if args.result_json:
            open(args.result_json, "w").write(json_dumps(result_status))
        finish_args = dict(
            client=client,
            job_completed_normally=True,
            cleanup_job=args.cleanup,
            client_outputs=client_outputs,
            pulsar_outputs=pulsar_outputs,
        )
        failed = finish_job(**finish_args)
        return failed
    finally:
        client_manager.shutdown()


if __name__ == "__main__":
    main()



from os import system
from os.path import (
    abspath,
    join,
)

from pulsar.core import DEFAULT_STAGING_DIRECTORY
from pulsar.main import (
    ArgumentParser,
    PulsarConfigBuilder,
)

DESCRIPTION = "Change ownership of a job working directory."
# Switch this to true to tighten up security somewhat in production mode,
# better increase of security can be had by simply restricting sudoers rule
# to only allow chown -R of directories of form ${staging_directory}/${job_id}
FORCE_PRODUCTION = False


def main(argv=None):
    arg_parser = ArgumentParser(description=DESCRIPTION)
    arg_parser.add_argument("--user", required=True)
    arg_parser.add_argument("--job_id")
    arg_parser.add_argument("--job_directory")
    args = arg_parser.parse_args(argv)
    user = args.user
    job_id = args.job_id

    if args.job_id:
        staging_directory = PulsarConfigBuilder().load().get('staging_directory', DEFAULT_STAGING_DIRECTORY)
        job_directory = abspath(join(staging_directory, job_id))
        assert job_directory.startswith(staging_directory)
    elif FORCE_PRODUCTION:
        raise Exception("In production mode, must specify a job_id instead of a working directory.")
    else:
        job_directory = abspath(args.job_directory)
        assert job_directory
    command = "chown -Rh '{}' '{}'".format(user, job_directory)
    system(command)


if __name__ == "__main__":
    main()



"""
Tests for the relay transport implementation.

Tests retry logic and message ID tracking functionality.
"""
from unittest import TestCase
from unittest.mock import Mock, patch

import requests

from pulsar.client.transport.relay import RelayTransport, RelayTransportError


class TestRetryLogic(TestCase):
    """Test retry logic with exponential backoff."""

    @patch('pulsar.client.transport.relay.time.sleep')
    def test_post_message_retries_on_connection_error(self, mock_sleep):
        """Test that post_message retries indefinitely on connection errors."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')

        # Mock the auth manager to return a token
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Mock session.post to fail twice with ConnectionError, then succeed
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            'message_id': 'msg_123',
            'topic': 'test-topic',
            'timestamp': '2025-10-27T10:00:00Z'
        }

        transport.session.post = Mock(
            side_effect=[
                requests.ConnectionError("Connection refused"),
                requests.ConnectionError("Connection refused"),
                mock_response
            ]
        )

        result = transport.post_message('test-topic', {'data': 'test'})

        # Verify it succeeded after retries
        assert result['message_id'] == 'msg_123'
        assert transport.session.post.call_count == 3
        # Verify exponential backoff was used
        assert mock_sleep.call_count == 2
        # First delay should be 1.0, second should be 2.0
        assert mock_sleep.call_args_list[0][0][0] == 1.0
        assert mock_sleep.call_args_list[1][0][0] == 2.0

    @patch('pulsar.client.transport.relay.time.sleep')
    def test_post_message_retries_on_500_error(self, mock_sleep):
        """Test that post_message retries on 5xx server errors."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Mock responses: 500, 503, then 200
        mock_500 = Mock()
        mock_500.status_code = 500

        mock_503 = Mock()
        mock_503.status_code = 503

        mock_200 = Mock()
        mock_200.status_code = 200
        mock_200.json.return_value = {
            'message_id': 'msg_456',
            'topic': 'test-topic',
            'timestamp': '2025-10-27T10:00:00Z'
        }

        transport.session.post = Mock(side_effect=[mock_500, mock_503, mock_200])

        result = transport.post_message('test-topic', {'data': 'test'})

        assert result['message_id'] == 'msg_456'
        assert transport.session.post.call_count == 3
        assert mock_sleep.call_count == 2

    def test_post_message_does_not_retry_on_400_error(self):
        """Test that post_message does NOT retry on 4xx client errors."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Mock response with 400 error
        mock_400 = Mock()
        mock_400.status_code = 400

        # Create HTTPError with response attached
        error = requests.HTTPError("400 Bad Request")
        error.response = mock_400
        mock_400.raise_for_status.side_effect = error

        transport.session.post = Mock(return_value=mock_400)

        with self.assertRaises(RelayTransportError):
            transport.post_message('test-topic', {'data': 'test'})

        # Should only be called once (no retries for 4xx)
        assert transport.session.post.call_count == 1

    @patch('pulsar.client.transport.relay.time.sleep')
    def test_post_message_retries_on_timeout(self, mock_sleep):
        """Test that post_message retries on timeout."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            'message_id': 'msg_789',
            'topic': 'test-topic',
            'timestamp': '2025-10-27T10:00:00Z'
        }

        transport.session.post = Mock(
            side_effect=[
                requests.Timeout("Request timed out"),
                mock_response
            ]
        )

        result = transport.post_message('test-topic', {'data': 'test'})

        assert result['message_id'] == 'msg_789'
        assert transport.session.post.call_count == 2
        assert mock_sleep.call_count == 1

    @patch('pulsar.client.transport.relay.time.sleep')
    def test_retry_backoff_caps_at_max_delay(self, mock_sleep):
        """Test that exponential backoff caps at max_delay."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Create many connection errors to test max delay
        errors = [requests.ConnectionError("Connection refused")] * 10

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            'message_id': 'msg_999',
            'topic': 'test-topic',
            'timestamp': '2025-10-27T10:00:00Z'
        }

        transport.session.post = Mock(side_effect=errors + [mock_response])

        result = transport.post_message('test-topic', {'data': 'test'})

        assert result['message_id'] == 'msg_999'
        assert mock_sleep.call_count == 10

        # Check that delay caps at 60 seconds
        delays = [call[0][0] for call in mock_sleep.call_args_list]
        # Expected: 1, 2, 4, 8, 16, 32, 60, 60, 60, 60
        assert delays[0] == 1.0
        assert delays[1] == 2.0
        assert delays[2] == 4.0
        assert delays[3] == 8.0
        assert delays[4] == 16.0
        assert delays[5] == 32.0
        # After this, should cap at 60
        assert all(d == 60.0 for d in delays[6:])


class TestMessageIDTracking:
    """Test message ID tracking functionality."""

    def test_long_poll_tracks_message_ids(self):
        """Test that long_poll tracks message IDs per topic."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Mock response with messages from different topics
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            'messages': [
                {'topic': 'topic1', 'message_id': 'msg_001', 'payload': {'data': 'a'}},
                {'topic': 'topic2', 'message_id': 'msg_002', 'payload': {'data': 'b'}},
                {'topic': 'topic1', 'message_id': 'msg_003', 'payload': {'data': 'c'}},
            ],
            'has_more': False
        }

        transport.session.post = Mock(return_value=mock_response)

        messages = transport.long_poll(['topic1', 'topic2'])

        # Verify message IDs are tracked (last message ID per topic)
        assert transport.get_last_message_id('topic1') == 'msg_003'
        assert transport.get_last_message_id('topic2') == 'msg_002'
        assert len(messages) == 3

    def test_long_poll_uses_tracked_message_ids_in_since(self):
        """Test that long_poll includes tracked message IDs in the since parameter."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Set some tracked message IDs
        transport.set_last_message_id('topic1', 'msg_100')
        transport.set_last_message_id('topic2', 'msg_200')

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            'messages': [],
            'has_more': False
        }

        transport.session.post = Mock(return_value=mock_response)

        # Call long_poll
        transport.long_poll(['topic1', 'topic2'])

        # Verify the 'since' parameter was included in the request
        call_args = transport.session.post.call_args
        request_json = call_args[1]['json']

        assert 'since' in request_json
        assert request_json['since']['topic1'] == 'msg_100'
        assert request_json['since']['topic2'] == 'msg_200'

    def test_long_poll_only_includes_since_for_requested_topics(self):
        """Test that since only includes tracked IDs for topics in the request."""
        transport = RelayTransport('http://localhost:8000', 'user', 'pass')
        transport.auth_manager.get_token = Mock(return_value='test-token')

        # Set tracked message IDs for multiple topics
        transport.set_last_message_id('topic1', 'msg_100')
        transport.set_last_message_id('topic2', 'msg_200')
        transport.set_last_message_id('topic3', 'msg_300')

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {'messages': [], 'has_more': False}

        transport.session.post = Mock(return_value=mock_response)

        # Only poll for topic1 and topic2
        transport.long_poll(['topic1', 'topic2'])

        call_args = transport.session.post.call_args
        request_json = call_args[1]['json']

        # Should only include topic1 and topic2 in since
        assert 'since' in request_json
        assert 'topic1' in request_json['since']
        assert 'topic2' in request_json['since']
        assert 'topic3' not in request_json['since']



import contextlib
import datetime
import os
import threading
import time
from functools import partial

try:
    # If galaxy-lib or Galaxy 19.05 present.
    from galaxy.tools.deps.dependencies import DependenciesDescription
except ImportError:
    # If galaxy-tool-util or Galaxy 19.09 present.
    from galaxy.tool_util.deps.dependencies import DependenciesDescription

import logging

from pulsar.client.util import filter_destination_params
from pulsar.managers import (
    ManagerProxy,
    status,
)
from pulsar.managers.util.retry import RetryActionExecutor
from .staging import (
    postprocess,
    preprocess,
)

log = logging.getLogger(__name__)

DEFAULT_DO_MONITOR = False

DECACTIVATE_FAILED_MESSAGE = "Failed to deactivate job with job id %s. May cause problems on next Pulsar start."
ACTIVATE_FAILED_MESSAGE = "Failed to activate job with job id %s. This job may not recover properly upon Pulsar restart."

JOB_FILE_FINAL_STATUS = "final_status"
JOB_FILE_POSTPROCESSED = "postprocessed"
JOB_FILE_PREPROCESSED = "preprocessed"
JOB_FILE_PREPROCESSING_FAILED = "preprocessing_failed"
JOB_METADATA_RUNNING = "running"

ACTIVE_STATUS_PREPROCESSING = "preprocessing"
ACTIVE_STATUS_LAUNCHED = "launched"

DEFAULT_MIN_POLLING_INTERVAL = 0.5


class StatefulManagerProxy(ManagerProxy):
    """
    """

    def __init__(self, manager, **manager_options):
        super().__init__(manager)
        min_polling_interval = float(manager_options.get("min_polling_interval", DEFAULT_MIN_POLLING_INTERVAL))
        preprocess_retry_action_kwds = filter_destination_params(manager_options, "preprocess_action_")
        postprocess_retry_action_kwds = filter_destination_params(manager_options, "postprocess_action_")
        self.__preprocess_action_executor = RetryActionExecutor(**preprocess_retry_action_kwds)
        self.__postprocess_action_executor = RetryActionExecutor(**postprocess_retry_action_kwds)
        self.min_polling_interval = datetime.timedelta(0, min_polling_interval)
        self.active_jobs = ActiveJobs.from_manager(manager)
        self.__state_change_callback = self._default_status_change_callback
        self.__monitor = None

    def set_state_change_callback(self, state_change_callback):
        self.__state_change_callback = state_change_callback
        self.__monitor = ManagerMonitor(self)

    def _default_status_change_callback(self, status, job_id):
        log.info("Status of job [{}] changed to [{}]. No callbacks enabled.".format(job_id, status))

    def trigger_state_change_callback(self, job_id):
        proxy_status = self.get_status(job_id)
        log.debug("Triggering state change callback with status %s by request for job_id %s", proxy_status, job_id)
        self.__state_change_callback(proxy_status, job_id)

    @property
    def name(self):
        return self._proxied_manager.name

    def setup_job(self, *args, **kwargs):
        job_id = self._proxied_manager.setup_job(*args, **kwargs)
        return job_id

    def _persist_launch_config(self, job_id, launch_config):
        job_directory = self._proxied_manager.job_directory(job_id)
        job_directory.store_metadata("launch_config", launch_config)

    def touch_outputs(self, job_id, touch_outputs):
        job_directory = self._proxied_manager.job_directory(job_id)
        for name in touch_outputs:
            path = job_directory.calculate_path(name, 'output')
            job_directory.open_file(path, mode='a')

    def preprocess_and_launch(self, job_id, launch_config):
        self._persist_launch_config(job_id, launch_config)
        requires_preprocessing = launch_config.get("remote_staging") and launch_config["remote_staging"].get("setup")
        if requires_preprocessing:
            self.active_jobs.activate_job(job_id, active_status=ACTIVE_STATUS_PREPROCESSING)
            self._launch_prepreprocessing_thread(job_id, launch_config)
        else:
            with self._handling_of_preprocessing_state(job_id, launch_config):
                pass

    def _launch_prepreprocessing_thread(self, job_id, launch_config):
        def do_preprocess():
            with self._handling_of_preprocessing_state(job_id, launch_config):
                job_directory = self._proxied_manager.job_directory(job_id)
                was_cancelled = partial(self._proxied_manager._was_cancelled, job_id)
                staging_config = launch_config.get("remote_staging", {})
                # TODO: swap out for a generic "job_extra_params"
                if 'action_mapper' in staging_config and \
                        'ssh_key' in staging_config['action_mapper'] and \
                        'setup' in staging_config:
                    for action in staging_config['setup']:
                        action['action'].update(ssh_key=staging_config['action_mapper']['ssh_key'])
                setup_config = staging_config.get("setup", [])
                preprocess(job_directory, setup_config, self.__preprocess_action_executor, was_cancelled, object_store=self.object_store)
                self.active_jobs.deactivate_job(job_id, active_status=ACTIVE_STATUS_PREPROCESSING)

        new_thread_for_job(self, "preprocess", job_id, do_preprocess, daemon=False)

    @contextlib.contextmanager
    def _handling_of_preprocessing_state(self, job_id, launch_config):
        job_directory = self._proxied_manager.job_directory(job_id)
        try:
            yield
            if self._proxied_manager._was_cancelled(job_id):
                log.info("Exiting job launch, job is cancelled")
                return
            launch_kwds = {}
            if launch_config.get("dependencies_description"):
                dependencies_description = DependenciesDescription.from_dict(launch_config["dependencies_description"])
                launch_kwds["dependencies_description"] = dependencies_description
            for kwd in ["submit_params", "setup_params", "env"]:
                if kwd in launch_config:
                    launch_kwds[kwd] = launch_config[kwd]

            self._proxied_manager.launch(
                job_id,
                launch_config["command_line"],
                **launch_kwds
            )
            with job_directory.lock("status"):
                job_directory.store_metadata(JOB_FILE_PREPROCESSED, True)
            self.active_jobs.activate_job(job_id)
        except Exception as e:
            with job_directory.lock("status"):
                job_directory.store_metadata(JOB_FILE_PREPROCESSING_FAILED, True)
                job_directory.store_metadata("return_code", 1)
                job_directory.write_file("stderr", str(e))
            self.__state_change_callback(status.FAILED, job_id)
            log.exception("Failed job preprocessing for job %s:", job_id)

    def handle_failure_before_launch(self, job_id):
        self.__state_change_callback(status.FAILED, job_id)

    def get_status(self, job_id):
        """ Compute status used proxied manager and handle state transitions
        and track additional state information needed.
        """
        job_directory = self._proxied_manager.job_directory(job_id)
        if not job_directory.exists():
            return status.LOST

        with job_directory.lock("status"):
            proxy_status, state_change = self.__proxy_status(job_directory, job_id)

        if state_change == "to_complete":
            self.__deactivate(job_id, proxy_status)
        elif state_change == "to_running":
            self.__state_change_callback(status.RUNNING, job_id)

        return self.__status(job_directory, proxy_status)

    def __proxy_status(self, job_directory, job_id):
        """ Determine state with proxied job manager and if this job needs
        to be marked as deactivated (this occurs when job first returns a
        complete status from proxy.
        """
        state_change = None
        if job_directory.has_metadata(JOB_FILE_PREPROCESSING_FAILED):
            proxy_status = status.FAILED
            job_directory.store_metadata(JOB_FILE_FINAL_STATUS, proxy_status)
            state_change = "to_complete"
        elif not job_directory.has_metadata(JOB_FILE_PREPROCESSED):
            proxy_status = status.PREPROCESSING
        elif job_directory.has_metadata(JOB_FILE_FINAL_STATUS):
            proxy_status = job_directory.load_metadata(JOB_FILE_FINAL_STATUS)
        else:
            proxy_status = self._proxied_manager.get_status(job_id)
            if proxy_status == status.RUNNING:
                if not job_directory.has_metadata(JOB_METADATA_RUNNING):
                    job_directory.store_metadata(JOB_METADATA_RUNNING, True)
                    state_change = "to_running"
            elif proxy_status in [status.COMPLETE, status.CANCELLED]:
                job_directory.store_metadata(JOB_FILE_FINAL_STATUS, proxy_status)
                state_change = "to_complete"
        return proxy_status, state_change

    def __status(self, job_directory, proxy_status):
        """ Use proxied manager's status to compute the real
        (stateful) status of job.
        """
        if proxy_status == status.COMPLETE:
            if not job_directory.has_metadata(JOB_FILE_POSTPROCESSED):
                job_status = status.POSTPROCESSING
            else:
                job_status = status.COMPLETE
        else:
            job_status = proxy_status
        return job_status

    def __deactivate(self, job_id, proxy_status):
        self.active_jobs.deactivate_job(job_id)
        deactivate_method = getattr(self._proxied_manager, "_deactivate_job", None)
        if deactivate_method:
            try:
                deactivate_method(job_id)
            except Exception:
                log.exception("Failed to deactivate via proxied manager job %s" % job_id)
        if proxy_status == status.COMPLETE:
            self.__handle_postprocessing(job_id)

    def __handle_postprocessing(self, job_id):
        def do_postprocess():
            postprocess_success = False
            job_directory = self._proxied_manager.job_directory(job_id)
            was_cancelled = partial(self._proxied_manager._was_cancelled, job_id)
            try:
                postprocess_success = postprocess(job_directory, self.__postprocess_action_executor, was_cancelled)
            except Exception:
                log.exception("Failed to postprocess results for job id %s" % job_id)
            final_status = status.COMPLETE if postprocess_success else status.FAILED
            if job_directory.has_metadata(JOB_FILE_PREPROCESSING_FAILED):
                final_status = status.FAILED
            self.__state_change_callback(final_status, job_id)
        new_thread_for_job(self, "postprocess", job_id, do_postprocess, daemon=False)

    def shutdown(self, timeout=None):
        if self.__monitor:
            try:
                self.__monitor.shutdown(timeout)
            except Exception:
                log.exception("Failed to shutdown job monitor for manager %s" % self.name)
        super().shutdown(timeout)

    def recover_active_jobs(self):
        unqueue_preprocessing_ids = []
        for job_id in self.active_jobs.active_job_ids(active_status=ACTIVE_STATUS_PREPROCESSING):
            job_directory = self._proxied_manager.job_directory(job_id)
            if not job_directory.has_metadata("launch_config"):
                log.warn("Failed to find launch parameters for job scheduled to prepreprocess [%s]" % job_id)
                unqueue_preprocessing_ids.append(job_id)
            elif job_directory.has_metadata(JOB_FILE_PREPROCESSED):
                log.warn("Job scheduled to prepreprocess [%s] already preprocessed, skipping" % job_id)
                unqueue_preprocessing_ids.append(job_id)
            elif job_directory.has_metadata(JOB_FILE_PREPROCESSING_FAILED):
                log.warn("Job scheduled to prepreprocess [%s] previously failed preprocessing, skipping" % job_id)
                unqueue_preprocessing_ids.append(job_id)
            else:
                launch_config = job_directory.load_metadata("launch_config")
                self._launch_prepreprocessing_thread(job_id, launch_config)

        for unqueue_preprocessing_id in unqueue_preprocessing_ids:
            self.active_jobs.deactivate_job(unqueue_preprocessing_id, active_status=ACTIVE_STATUS_PREPROCESSING)

        recover_method = getattr(self._proxied_manager, "_recover_active_job", None)
        if recover_method is None:
            return

        for job_id in self.active_jobs.active_job_ids(active_status=ACTIVE_STATUS_LAUNCHED):
            try:
                recover_method(job_id)
            except Exception:
                log.exception("Failed to recover active job %s" % job_id)
                self.__handle_recovery_problem(job_id)

    def __handle_recovery_problem(self, job_id):
        # Make sure we tell the client we have lost this job.
        self.active_jobs.deactivate_job(job_id)
        self.__state_change_callback(status.LOST, job_id)


class ActiveJobs:
    """ Keeps track of active jobs (those that are not yet "complete").
    Current implementation is file based, but could easily be made
    database-based instead.

    TODO: Keep jobs in memory after initial load so don't need to repeatedly
    hit disk to recover this information.
    """

    @staticmethod
    def from_manager(manager):
        persistence_directory = manager.persistence_directory
        manager_name = manager.name
        return ActiveJobs(manager_name, persistence_directory)

    def __init__(self, manager_name, persistence_directory):
        if persistence_directory:
            active_job_directory = os.path.join(persistence_directory, "%s-active-jobs" % manager_name)
            if not os.path.exists(active_job_directory):
                os.makedirs(active_job_directory)
            preprocessing_job_directory = os.path.join(persistence_directory, "%s-preprocessing-jobs" % manager_name)
            if not os.path.exists(preprocessing_job_directory):
                os.makedirs(preprocessing_job_directory)
        else:
            active_job_directory = None
            preprocessing_job_directory = None
        self.launched_job_directory = active_job_directory
        self.preprocessing_job_directory = preprocessing_job_directory

    def active_job_ids(self, active_status=ACTIVE_STATUS_LAUNCHED):
        job_ids = []
        target_directory = self._active_job_directory(active_status)
        if target_directory:
            job_ids = os.listdir(target_directory)
        return job_ids

    def activate_job(self, job_id, active_status=ACTIVE_STATUS_LAUNCHED):
        if self._active_job_directory(active_status):
            path = self._active_job_file(job_id, active_status=active_status)
            try:
                open(path, "w").close()
            except Exception:
                log.warn(ACTIVATE_FAILED_MESSAGE % job_id)

    def deactivate_job(self, job_id, active_status=ACTIVE_STATUS_LAUNCHED):
        if self._active_job_directory(active_status):
            path = self._active_job_file(job_id, active_status=active_status)
            if os.path.exists(path):
                try:
                    os.remove(path)
                except Exception:
                    log.warn(DECACTIVATE_FAILED_MESSAGE % job_id)

    def _active_job_directory(self, active_status):
        if active_status == ACTIVE_STATUS_LAUNCHED:
            target_directory = self.launched_job_directory
        elif active_status == ACTIVE_STATUS_PREPROCESSING:
            target_directory = self.preprocessing_job_directory
        else:
            raise Exception("Unknown active state encountered [%s]" % active_status)
        return target_directory

    def _active_job_file(self, job_id, active_status=ACTIVE_STATUS_LAUNCHED):
        return os.path.join(self._active_job_directory(active_status), job_id)


class ManagerMonitor:
    """ Monitors active jobs of a StatefulManagerProxy.
    """

    def __init__(self, stateful_manager):
        self.stateful_manager = stateful_manager
        self.active = True
        thread = new_thread_for_manager(self.stateful_manager, "[action=monitor]", self._run, True)
        self.thread = thread

    def shutdown(self, timeout=None):
        self.active = False
        self.thread.join(timeout)
        if self.thread.is_alive():
            log.warn("Failed to join monitor thread [%s]" % self.thread)

    def _run(self):
        """ Main loop, repeatedly checking active jobs of stateful manager.
        """
        while self.active:
            try:
                self._monitor_active_jobs()
            except Exception:
                log.exception("Failure in stateful manager monitor step.")
                # This should hopefully be a rare event.
                # Let's not hammer the system with job lookups
                time.sleep(1)

    def _monitor_active_jobs(self):
        active_job_ids = self.stateful_manager.active_jobs.active_job_ids()
        iteration_start = datetime.datetime.now()
        for active_job_id in active_job_ids:
            try:
                self._check_active_job_status(active_job_id)
            except Exception:
                log.exception("Failed checking active job status for job_id %s" % active_job_id)
        iteration_end = datetime.datetime.now()
        iteration_length = iteration_end - iteration_start
        if iteration_length < self.stateful_manager.min_polling_interval:
            to_sleep = (self.stateful_manager.min_polling_interval - iteration_length)
            microseconds = to_sleep.microseconds + (to_sleep.seconds + to_sleep.days * 24 * 3600) * (10 ** 6)
            total_seconds = microseconds / (10 ** 6)
            time.sleep(total_seconds)

    def _check_active_job_status(self, active_job_id):
        # Manager itself will handle state transitions when status changes,
        # just need to poll get_status
        self.stateful_manager.get_status(active_job_id)


def new_thread_for_job(manager, action, job_id, target, daemon):
    name = "[action={}]-[job={}]".format(action, job_id)
    return new_thread_for_manager(manager, name, target, daemon)


def new_thread_for_manager(manager, name, target, daemon):
    thread_name = "[manager={}]-{}".format(manager.name, name)
    thread = threading.Thread(name=thread_name, target=target)
    thread.daemon = daemon
    thread.start()
    return thread


__all__ = ('StatefulManagerProxy',)



import logging
import os
import stat

from galaxy.util import asbool

from pulsar.managers import PULSAR_UNKNOWN_RETURN_CODE
from pulsar.managers.base import BaseManager
from ..util.env import env_to_statement
from ..util.job_script import job_script

log = logging.getLogger(__name__)

# TODO: Rename these to abstract out the fact they are files - pulsar
# should be able to replace metadata backing with non-file stuff now that
# the abstractions are fairly well utilized.
JOB_FILE_RETURN_CODE = "return_code"
TOOL_FILE_STANDARD_OUTPUT = os.path.join("metadata", "tool_stdout")
TOOL_FILE_STANDARD_ERROR = os.path.join("metadata", "tool_stderr")
JOB_FILE_STANDARD_OUTPUT = os.path.join("metadata", "job_stdout")
JOB_FILE_STANDARD_ERROR = os.path.join("metadata", "job_stderr")
JOB_FILE_TOOL_ID = "tool_id"
JOB_FILE_TOOL_VERSION = "tool_version"
JOB_FILE_CANCELLED = "cancelled"
JOB_FILE_COMMAND_LINE = "command_line"
CREATE_TMP_PATTERN = '''$([ ! -e '{0}/tmp' ] || mv '{0}/tmp' '{0}'/tmp.$(date +%Y%m%d-%H%M%S) ; mkdir '{0}/tmp'; echo '{0}/tmp')'''
PREPARE_DIRS_TEMPLATE = """for dir in {working_directory} {outputs_directory} {configs_directory}; do
    if [ -d "${{dir}}.orig" ]; then
        rm -rf "$dir"; cp -R "${{dir}}.orig" "$dir"
    else
        cp -R "$dir" "${{dir}}.orig"
    fi
done
"""


class DirectoryBaseManager(BaseManager):

    def _job_file(self, job_id, name):
        return self._job_directory(job_id)._job_file(name)

    def return_code(self, job_id):
        return_code_str = self._read_job_file(job_id, JOB_FILE_RETURN_CODE, default=PULSAR_UNKNOWN_RETURN_CODE)
        return int(return_code_str) if return_code_str and return_code_str != PULSAR_UNKNOWN_RETURN_CODE else return_code_str

    def stdout_contents(self, job_id):
        try:
            return self._read_job_file(job_id, TOOL_FILE_STANDARD_OUTPUT, size=self.maximum_stream_size)
        except FileNotFoundError:
            # Could be old job finishing up, drop in 2024?
            return self._read_job_file(job_id, "tool_stdout", size=self.maximum_stream_size, default=b"")

    def stderr_contents(self, job_id):
        try:
            return self._read_job_file(job_id, TOOL_FILE_STANDARD_ERROR, size=self.maximum_stream_size)
        except FileNotFoundError:
            # Could be old job finishing up, drop in 2024?
            return self._read_job_file(job_id, "tool_stderr", size=self.maximum_stream_size, default=b"")

    def job_stdout_contents(self, job_id):
        return self._read_job_file(job_id, JOB_FILE_STANDARD_OUTPUT, size=self.maximum_stream_size, default=b"")

    def job_stderr_contents(self, job_id):
        return self._read_job_file(job_id, JOB_FILE_STANDARD_ERROR, size=self.maximum_stream_size, default=b"")

    def read_command_line(self, job_id):
        command_line = self._read_job_file(job_id, JOB_FILE_COMMAND_LINE)
        if command_line.startswith(b'"'):
            # legacy JSON...
            import json
            command_line = json.loads(command_line)
        return command_line

    def _tool_stdout_path(self, job_id):
        return self._job_file(job_id, TOOL_FILE_STANDARD_OUTPUT)

    def _tool_stderr_path(self, job_id):
        return self._job_file(job_id, TOOL_FILE_STANDARD_ERROR)

    def _job_stdout_path(self, job_id):
        return self._job_file(job_id, JOB_FILE_STANDARD_OUTPUT)

    def _job_stderr_path(self, job_id):
        return self._job_file(job_id, JOB_FILE_STANDARD_ERROR)

    def _return_code_path(self, job_id):
        return self._job_file(job_id, JOB_FILE_RETURN_CODE)

    def _setup_job_for_job_id(self, job_id, tool_id, tool_version):
        self._setup_job_directory(job_id)

        tool_id = str(tool_id) if tool_id else ""
        tool_version = str(tool_version) if tool_version else ""

        authorization = self._get_authorization(job_id, tool_id)
        authorization.authorize_setup()

        self._write_tool_info(job_id, tool_id, tool_version)
        return job_id

    def _read_job_file(self, job_id, name, **kwds):
        return self._job_directory(job_id).read_file(name, **kwds)

    def _write_job_file(self, job_id, name, contents):
        return self._job_directory(job_id).write_file(name, contents)

    def _write_return_code_if_unset(self, job_id, return_code):
        return_code_str = self._read_job_file(job_id, JOB_FILE_RETURN_CODE, default=PULSAR_UNKNOWN_RETURN_CODE)
        if return_code_str == PULSAR_UNKNOWN_RETURN_CODE:
            self._write_job_file(job_id, JOB_FILE_RETURN_CODE, str(return_code))

    def _write_tool_info(self, job_id, tool_id, tool_version):
        job_directory = self._job_directory(job_id)
        job_directory.store_metadata(JOB_FILE_TOOL_ID, tool_id)
        job_directory.store_metadata(JOB_FILE_TOOL_VERSION, tool_version)

    def _write_command_line(self, job_id, command_line):
        self._write_job_file(job_id, JOB_FILE_COMMAND_LINE, command_line)

    def _record_cancel(self, job_id):
        try:
            self._job_directory(job_id).store_metadata(JOB_FILE_CANCELLED, True)
        except Exception:
            log.info("Failed to record job with id %s was cancelled." % job_id)

    def _was_cancelled(self, job_id):
        try:
            return self._job_directory(job_id).load_metadata(JOB_FILE_CANCELLED, None)
        except Exception:
            log.info("Failed to determine if job with id %s was cancelled, assuming no." % job_id)
            return False

    def _open_job_standard_output(self, job_id):
        return self._job_directory(job_id).open_file(JOB_FILE_STANDARD_OUTPUT, 'w')

    def _open_job_standard_error(self, job_id):
        return self._job_directory(job_id).open_file(JOB_FILE_STANDARD_ERROR, 'w')

    def _check_execution_with_tool_file(self, job_id, command_line):
        tool_id = self._tool_id(job_id)
        self._check_execution(job_id, tool_id, command_line)

    def _tool_id(self, job_id):
        tool_id = None
        job_directory = self._job_directory(job_id)
        if job_directory.has_metadata(JOB_FILE_TOOL_ID):
            tool_id = job_directory.load_metadata(JOB_FILE_TOOL_ID)
        return tool_id

    def _expand_command_line(self, job_id, command_line: str, dependencies_description, job_directory=None) -> str:
        command_line = super()._expand_command_line(
            job_id, command_line, dependencies_description, job_directory=job_directory
        )
        if not self._is_windows:
            rc_path = self._return_code_path(job_id)
            CAPTURE_RETURN_CODE = "return_code=$?"
            command_line = f"{command_line}; {CAPTURE_RETURN_CODE}; echo $return_code > {rc_path};"
        return command_line

    # Helpers methods related to setting up job script files.
    def _setup_job_file(self, job_id, command_line, dependencies_description=None, env=[], setup_params=None):
        command_line = self._expand_command_line(
            job_id, command_line, dependencies_description, job_directory=self.job_directory(job_id).job_directory
        )
        script_env = self._job_template_env(job_id, command_line=command_line, env=env, setup_params=setup_params)
        script = job_script(**script_env)
        return self._write_job_script(job_id, script)

    def _tmp_dir(self, job_id: str):
        # Code stolen from Galaxy's job wrapper.
        tmp_dir = self.tmp_dir
        try:
            if not tmp_dir or asbool(tmp_dir):
                working_directory = self.job_directory(job_id).job_directory
                return CREATE_TMP_PATTERN.format(working_directory)
            else:
                return tmp_dir
        except ValueError:
            # Catch case where tmp_dir is a complex expression and not a boolean value
            return tmp_dir

    def _prepare_dirs(self, job_id: str):
        return PREPARE_DIRS_TEMPLATE.format(
            working_directory=self.job_directory(job_id).working_directory(),
            outputs_directory=self.job_directory(job_id).outputs_directory(),
            configs_directory=self.job_directory(job_id).configs_directory(),
        )

    def _job_template_env(self, job_id, command_line=None, env=[], setup_params=None):
        # TODO: Add option to ignore remote env.
        env = env + self.env_vars
        setup_params = setup_params or {}
        env_setup_commands = map(env_to_statement, env)
        job_template_env = {
            'job_instrumenter': self.job_metrics.default_job_instrumenter,
            'galaxy_virtual_env': self._galaxy_virtual_env(),
            'galaxy_lib': self._galaxy_lib(),
            'preserve_python_environment': setup_params.get('preserve_galaxy_python_environment', False),
            'env_setup_commands': env_setup_commands,
            # job_diredctory not used by job_script and it calls the job directory working directory
            'working_directory': self.job_directory(job_id).working_directory(),
            'metadata_directory': self.job_directory(job_id).metadata_directory(),
            'home_directory': self.job_directory(job_id).home_directory(),
            'job_id': job_id,
            'tmp_dir_creation_statement': self._tmp_dir(job_id),
            'prepare_dirs_statement': self._prepare_dirs(job_id),
        }
        if command_line:
            job_template_env['command'] = command_line
        return job_template_env

    def _write_job_script(self, job_id, contents):
        self._write_job_file(job_id, "command.sh", contents)
        script_path = self._job_file(job_id, "command.sh")
        os.chmod(script_path, stat.S_IEXEC | stat.S_IWRITE | stat.S_IREAD)
        return script_path
